from enum import Enum
import subprocess


class SolverType(Enum):
    TRANSIENT_ANALYSIS = "transientAnalysis"
    TRANSIENT_SIMULATION = "transientSimulation"
    STATIONARY_ANALYSIS = "stationaryAnalysis"
    STATIONARY_SIMULATION = "stationarySimulation"


class ModelType(Enum):
    RBD = "RBD"
    SPN = "SPN"
    CTMC = "CTMC"
    DTMC = "DTMC"
    ET = "ET"
    FT = "FT"


class SPN:

    class ServerType(Enum):
        SINGLE_SERVER = "ExclusiveServer"
        INFINITE_SERVER = "InfiniteServer"

    class TransitionType(Enum):
        IMMEDIATE_TRANSITION = "immediateTransition"
        TIMED_TRANSITION = "timedTransition"

    class Place:

        def __init__(self, name, tokens=None):
            self.name = name
            self.tokens = tokens

    class __Transition:

        def __init__(self, name, inputs=None, outputs=None, priority=None, guard=None):
            self.name = name
            self.inputs = inputs if inputs is not None else []
            self.outputs = outputs if outputs is not None else []
            self.priority = priority
            self.guard = guard

    class ImmediateTransition(__Transition):

        def __init__(self, name, inputs=None, outputs=None, weight=1, priority=None, guard=None):
            super().__init__(name, inputs, outputs, priority, guard)
            self.weight = weight

    class TimedTransition(__Transition):

        def __init__(
            self,
            name,
            inputs=None,
            outputs=None,
            priority=None,
            guard=None,
            server_type: 'SPN.ServerType' = None,
            distribution=None,
            distribution_params=None,
        ):
            super().__init__(name, inputs, outputs, priority, guard)
            self.server_type = server_type
            self.distribution = distribution
            self.distribution_params = distribution_params

    class Model:

        def __init__(self):
            self.places = []
            self.transitions = []
            self.metrics = []
            self.solvers = []

        def add_place(self, place: 'SPN.Place'):
            self.places.append(place)

        def add_transition(self, transition: 'SPN.__Transition'):
            self.transitions.append(transition)

        def add_solver(self, solver: SolverType, metric_name, parameters=None):
            self.solvers.append((solver.value, metric_name, parameters))

        def set_metric(self, metric_name, metric_expression):
            self.metrics.append((metric_name, metric_expression))


class Mercury:

    def __init__(self, java_path=None, mercury_jar_path=None):
        self.int_vars = {}
        self.real_vars = {}
        self.models = {model_type.value: [] for model_type in ModelType}
        self.solvers = []
        self.java_path = java_path
        self.mercury_jar_path = mercury_jar_path

    def set_int_var(self, name, value):
        self.int_vars[name] = value

    def set_real_var(self, name, value):
        self.real_vars[name] = value

    def add_model(self, model_type: ModelType, model_instance, model_name):
        self.models[model_type.value].append((model_name, model_instance))

    def __generate_script(self):
        script_parts = []

        for name, value in self.real_vars.items():
            script_parts.append(f"{name} = {value};")
        for name, value in self.int_vars.items():
            script_parts.append(f"{name} = {value};")

        for model_type_str, models in self.models.items():
            for model_name, model in models:
                model_script = [f"\n{model_type_str} {model_name}{{"]
                if model_type_str == ModelType.SPN.value:
                    for place in model.places:
                        if place.tokens is not None:
                            model_script.append(f"\tplace {place.name}( tokens = {place.tokens} );")
                        else:
                            model_script.append(f"\tplace {place.name};")
                    for transition in model.transitions:
                        if isinstance(transition, SPN.ImmediateTransition):
                            transition_type = SPN.TransitionType.IMMEDIATE_TRANSITION.value
                        elif isinstance(transition, SPN.TimedTransition):
                            transition_type = SPN.TransitionType.TIMED_TRANSITION.value
                        else:
                            transition_type = "transition"
                        trans_script = [f"\t{transition_type} {transition.name}("]
                        params = []
                        if transition.inputs:
                            inputs = ', '.join([f"{input_tuple[0].name}({input_tuple[1]})" if len(input_tuple) > 1 and input_tuple[1] is not None else f"{input_tuple[0].name}" for input_tuple in transition.inputs])
                            params.append(f"inputs = [{inputs}]")
                        if transition.outputs:
                            outputs = ', '.join([f"{output_tuple[0].name}({output_tuple[1]})" if len(output_tuple) > 1 and output_tuple[1] is not None else f"{output_tuple[0].name}" for output_tuple in transition.outputs])
                            params.append(f"outputs = [{outputs}]")
                        if transition.priority is not None:
                            params.append(f"priority = {transition.priority}")
                        if transition.guard is not None:
                            params.append(f"guard = {transition.guard}")
                        if isinstance(transition, SPN.TimedTransition):
                            if transition.distribution is not None and transition.distribution_params is not None:
                                params_str = ', '.join(f"{k} = {v}" for k, v in transition.distribution_params.items())
                                delay_script = f"delay = ( type = \"{transition.distribution}\", parameters = ({params_str}) )"
                                params.append(delay_script)
                            if transition.server_type is not None:
                                server_type = transition.server_type.value
                                params.append(f"serverType = \"{server_type}\"")
                        if isinstance(transition, SPN.ImmediateTransition):
                            if transition.weight != 1:
                                params.append(f"weight = {transition.weight}")
                        trans_script.append('\t\t' + ',\n\t\t'.join(params))
                        trans_script.append("\t);")
                        model_script.extend(trans_script)
                    for solver, metric_name, parameters in model.solvers:
                        metric_expression = next((metric for metric in model.metrics if metric[0] == metric_name), None)
                        if metric_expression and parameters:
                            params_str = ',\n\t\t\t'.join(f"{k} = {v}" for k, v in parameters.items())
                            model_script.append(f"\n\tmetric {metric_name} = {solver}(\n\t\texpression = {metric_expression[1]},\n\t\tparameters = (\n\t\t\t{params_str}\n\t\t)\n\t);")
                    model_script.append("}\n")
                    script_parts.append('\n'.join(model_script))
                else:
                    pass

        main_script = ["main{"]
        if self.int_vars:
            keys = [f'\"{key}\"' for key in self.int_vars.keys()]
            main_script.append(f"\tsetIntegerParameters({', '.join(keys)});")
        for model_type_str, models in self.models.items():
            for model_name, model in models:
                if model.solvers:
                    for solver, metric_name, parameters in model.solvers:
                        main_script.append(f"\t{metric_name} = solve({model_name}, {metric_name});")
                        main_script.append(f"\tprintln({metric_name});")
        main_script.append("}")
        script_parts.append('\n'.join(main_script))

        return '\n'.join(script_parts)

    def run(self, script_path):
        print("Writing Mercury's script...")
        script = self.__generate_script()
        # print("start----------------------------------------------------------------")
        # print(script)
        # print("end----------------------------------------------------------------")
        print("The script has been written!")
        print("Saving script file...")
        with open(script_path, "w") as f:
            f.write(script)
        print("The script file has been saved!")
        command = [self.java_path, '-jar', self.mercury_jar_path, '-cli', script_path]
        print("Running Mercury's script...")
        try:
            result = subprocess.run(
                command,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            print(result.stdout)
        except subprocess.CalledProcessError as e:
            print(f"Command failed with error:\n{e.stderr}")
        except Exception as e:
            print(f"An unexpected error occurred: {e}")



import itertools
import math
import os
import pandas as pd
import numpy as np
import time
from VSCode.battery_mercury_experiments.a002.mercupy.mercupy import *

# 6 5 1 0   2S2P
# 9 5 1 0   3S2P
# 72 5 6 1  20S7P

java_path = r'C:\Program Files\Java\jdk-1.8\bin\java.exe'
jar_mercury_path = r'X:\BK\UFPE\Mestrado\Mercury\trunk\dist\Mercury.jar'


def clear_terminal():
    if os.name == 'nt':
        os.system('cls')
    else:
        os.system('clear')


clear_terminal()

v_bat = 72
i_bat = 20
t_bat = 1
C_bat = i_bat * t_bat

cells = []
with open(r'X:\BK\Projetos\Projeto_Mestrado\VSCode\battery_mercury_experiments\lithium_cells.txt', 'r') as file:
    for line in file:
        parts = line.strip().split(',')
        cells.append((parts[0], float(parts[1]), float(parts[2]), float(parts[3])))

    file.close()

df_cells = pd.DataFrame(cells, columns=['Battery Name', 'Nominal Voltage (V)', 'Continuous Discharge Current (A)', 'Capacity (mAh)'])
df_cells.index.name = 'Index'

print(df_cells)

selected_cell = df_cells.iloc[0]

print(f"Selected cell: {selected_cell['Battery Name']}")

print("-----------------------------------------------------------------------")
print("Results")
print("-----------------------------------------------------------------------")

print()

n_cells_series = math.ceil(v_bat / selected_cell['Nominal Voltage (V)'])
print(f"Number of cells in series: {n_cells_series}")

n_cells_parallel = math.ceil(C_bat / (selected_cell['Capacity (mAh)'] * 1e-3))
print(f"Number of cells in parallel: {n_cells_parallel}")

n_cells = n_cells_series * n_cells_parallel
print(f"Total number of cells: {n_cells}")

print()

print("Creating Mercury's Script...")

MNCD1PSOH_all = {
    "Group_1": (10.9032549732308, 3),
    "Group_2": (7.668014478961169, 3),
    "Group_3": (2.2271136238699802, 6),
    "Group_4": (2.135928112695071, 4),
    "Group_5": (1.2261594237806654, 5),
}

for group_name, (mean_delay, shape) in MNCD1PSOH_all.items():
    start_time = time.time()

    script = Mercury(
        java_path=java_path,
        mercury_jar_path=jar_mercury_path,
    )

    class INT_Vars(Enum):
        SOH_MIN = 0
        SOH_MAX = 20
        MNCD1PSOH_shape = round(shape)

    class REAL_Vars(Enum):
        MNCD1PSOH_mean_delay = mean_delay

    for var in INT_Vars:
        script.set_int_var(var.name, var.value)
    for var in REAL_Vars:
        script.set_real_var(var.name, var.value)

    cell_counter = itertools.count(start=0)
    transition_counter = itertools.count(start=0)

    battery_20S7P_SP = SPN.Model()

    place_cells_initial = [SPN.Place(name=f"P{next(cell_counter)}", tokens=INT_Vars.SOH_MAX.name) for _ in range(n_cells)]
    places_series_soh_deposit = [SPN.Place(name=f"P{next(cell_counter)}") for _ in range(n_cells)]
    places_series_ending = [SPN.Place(name=f"P{next(cell_counter)}") for _ in range(n_cells_parallel)]
    places_parallel_ending = [SPN.Place(name=f"P{next(cell_counter)}") for _ in range(n_cells_parallel)]
    place_battery_failure = SPN.Place(name=f"P{next(cell_counter)}")

    for place in [
            *place_cells_initial,
            *places_series_soh_deposit,
            *places_series_ending,
            *places_parallel_ending,
            place_battery_failure,
    ]:
        battery_20S7P_SP.add_place(place)

    immediate_transitions_series = [SPN.ImmediateTransition(
        name=f"TI{next(transition_counter)}",
        inputs=[(places_series_soh_deposit[i], f'{INT_Vars.SOH_MAX.name} - {INT_Vars.SOH_MIN.name}')],
        outputs=[(places_series_ending[i // n_cells_series], )],
    ) for i in range(n_cells)]

    immediate_transitions_parallel = [SPN.ImmediateTransition(
        name=f"TI{next(transition_counter)}",
        inputs=[(places_series_ending[i], )] + [(p, f'#{p.name}') for p in [*place_cells_initial[i * n_cells_series:i * n_cells_series + n_cells_series], *places_series_soh_deposit[i * n_cells_series:i * n_cells_series + n_cells_series]]],
        outputs=[(places_parallel_ending[i], )],
    ) for i in range(n_cells_parallel)]

    immediate_transition_failure = SPN.ImmediateTransition(
        name=f"TI{next(transition_counter)}",
        inputs=[(p, ) for p in places_parallel_ending],
        outputs=[(place_battery_failure, )],
    )

    timed_transition_soh_decay = [SPN.TimedTransition(
        name=f"TE{next(transition_counter)}",
        inputs=[(i, )],
        outputs=[(o, )],
        server_type=SPN.ServerType.SINGLE_SERVER,
        distribution='Erlang',
        distribution_params={
            "Mean_delay": REAL_Vars.MNCD1PSOH_mean_delay.name,
            "Shapes": INT_Vars.MNCD1PSOH_shape.name,
        },
    ) for i, o in zip(place_cells_initial, places_series_soh_deposit)]

    for transition in [
            *immediate_transitions_series,
            *immediate_transitions_parallel,
            immediate_transition_failure,
            *timed_transition_soh_decay,
    ]:
        battery_20S7P_SP.add_transition(transition)

    battery_20S7P_SP.set_metric(metric_name='UR', metric_expression=f'\"P{{#{place_battery_failure.name}>0}}\"')

    file_path = rf'X:/BK/Projetos/Projeto_Mestrado/VSCode/battery_mercury_experiments/a002/results/20S7P_sp_group_{group_name}'

    battery_20S7P_SP.add_solver(
        solver=SolverType.TRANSIENT_SIMULATION,
        metric_name='UR',
        parameters={
            "time": 1000,
            "confidenceLevel": .95,
            "maxRelativeError": .1,
            "samplingPoints": 100,
            "runs": 30,
            "replications": 50,
            "file": f'"{file_path}"',
        },
    )

    script.add_model(
        model_type=ModelType.SPN,
        model_name='battery',
        model_instance=battery_20S7P_SP,
    )

    script.run(script_path=f'{file_path}.txt')

    end_time = time.time()
    elapsed_time = end_time - start_time
    print(f"Elapsed time Group {group_name}: {elapsed_time} seconds")
